#include <stddef.h>
#include <stdint.h>

#include <arch/types.h>
#include <os9/const.h>

#include "../features.h"

#define OS9_H

#ifndef NDEBUG
  #define dbg kprint
#else
  #define dbg
#endif

#define NR_TASKS 16
#define NR_WINS  32

#define NR_MPOINTS 4
#define NR_CHANS  20

#define KiB 1024u
#define MiB (1024u * KiB)
#define GiB (1024ul * MiB)

#define PAGESIZE (4 * KiB)
#define ENABLE   1
#define DISABLE  0

#define QID(t, v, p) ((Qid) {.type = (t), .vers = (v), .path = (p)})

#define SEEK_SET 0
#define SEEK_CUR 1
#define SEEK_END 2
#define EOF -1

typedef struct context Context;
typedef struct proc Proc;
typedef struct task Task;
typedef struct win Win;
typedef struct chan Chan;
typedef struct qid Qid;
typedef struct dir Dir;
typedef struct mach Mach;
typedef struct map Map;
typedef struct nspace Nspace;
typedef struct fdset Fdset;
typedef struct ref Ref;

enum rforkflags {
	RFPROC = 1 << 0,
	RFNAMEG = 1 << 1,
	RFCNAMEG = 1 << 2,
	RFFDG = 1 <<  3,
	RFCFDG = 1 << 4,
	RFMEM = 1 << 5,
};

enum barrier_type {
	CODE,
	DATA,
};

enum devflags {
	O_READ   = 1 << 0,
	O_WRITE  = 1 << 1,
	O_RDWR   = 1 << 2,
	O_BIND   = 1 << 3,
	O_DIR    = 1 << 4,
	O_STAT   = 1 << 5,
};

enum pstates {
	PSYSTEM = 1 << 0,
};

enum pmodes {
	PDISABLED = 'D',
	PCOLD_START = 'C',
	PWARM_START = 'W',
	PIDLE = 'I',
	PNORMAL = 'N',
};

enum tflags {
	TLOCK_PREEMP = 1 << 0,
	TENABLED = 1 << 1,
};

enum tstates {
	TDORMANT,
	TREADY,
	TRUNNING,
	TWAITING,
	FAULTED,
};

enum start_reason {
	SNORMAL,
};

enum map_attr {
	MR = 1 << 0,
	MW = 1 << 1,
	MX = 1 << 2,
	MD = 1 << 3,
};

struct ref {
	mutex_t m;
	int cnt;
};

struct map {
	Ref ref;
	char *name;
	unsigned perm;

	phyaddr_t pa;
	uintptr_t va;
	size_t siz;
};

/*
 * phystack and ktzero must be the first two fields of Mach
 * because crt.s expects that.
 */
struct mach {
	phyaddr_t phystack;
	phyaddr_t phytext;
	phyaddr_t kzero;

	Map *maps;
};

struct qid {
	unsigned long long path;
	unsigned long vers;
	unsigned char type;
};

struct dir {
	char name[NAMELEN];
	long length;
	unsigned char mode;
	unsigned char type;
	unsigned char dev;
	Qid qid;
};

struct chan {
	long offset;
	int index;
	Qid qid;
	unsigned char type;
	unsigned char dev;
	unsigned char mode;
	mutex_t mutex;
};

struct mpoint {
	Chan *new;
	Chan *old;
	mutex_t m;
};

struct task {
	int tid;
	int pid;

	char *name;
	void *entry;

	int prio;
	int baseprio;
	int retainprio;

	int wait;
	int locklevel;
	unsigned long affinity;

	long period;
	long duration;
	long capacity;

	long deadline;
	long deadtime;

	int state;
	int flags;

	char mode;
	int start;

	mutex_t m;

	Context ctx;
	/* communication ports */
	/* healt monitor table */
	/* Task *errhdl; */

	Map *text;
	Map *data;
	Map *stack;
	Nspace *ns;
	Fdset *fds;
};

struct win {
	int wid;
	long offset;
	long duration;
	int pid;
};

/* alloc.c */
extern void *alloc(size_t size);
extern void *allocb(int);

/* init.c */
extern void isys(void);

/* dlang.c */
extern int debug(void);

/* arch functions */
extern Context *getctx(Context *ctx);
extern void interrupt(int);
extern uint8_t inm8(void *addr);
extern uint16_t inm16(void *addr);
extern uint32_t inm32(void *addr);
extern uint8_t outm8(uint8_t, void *addr);
extern uint16_t outm16(uint16_t, void *addr);
extern uint32_t outm32(uint32_t, void *addr);
extern void lock(mutex_t *m);
extern void unlock(mutex_t *m);
extern int trylock(mutex_t *m);
extern void barrier(int);
extern noreturn void halt(void);
extern noreturn void panic(const char *msg);
extern noreturn void fault(const char *msg, Context *ctx);
extern noreturn void swtch(Context *ctx);
extern noreturn void trap(Context *ctx);

/* dev functions */
extern void idev(void);
extern Chan *devopen(char *name, int mode);
extern int devwrite(Chan *c, void *buf, int n);
extern int devread(Chan *c, void *buf, int n);
extern int devclose(Chan *c);

/* sched.c */
extern int getntask(int n, Task **tpp);
extern Task *gettask(int tid);
extern void sched(void);

/* globals */
extern Chan *console;

/* per cpu globals */
extern Proc *proc;
extern Task *task;
extern int cpuid;
extern long long now;
