#include <errno.h>
#include <stdarg.h>
#include <stddef.h>
#include <string.h>

#include <libk.h>

#include "stream.h"

static void
printc(Stream *sp, int c)
{
	if (sp->base) {
		if (sp->cnt < sp->len)
			sp->base[sp->cnt++] = c;
	} else {
		putch(c);
		sp->cnt++;
	}
}

static void
printn(Stream *sp, long long n, int base, int sign, int fill)
{
	int first, d;
	unsigned long long div;
	const static char digits[] = "0123456789ABCDEF";

	switch (base) {
	case 8:
		div = 01000000000000000000000u;
		break;
	case 10:
		div = 10000000000000000000u;
		break;
	case 16:
	default:
		div = 0x1000000000000000u;
		break;
	}

	if (sign && n < 0) {
		n = -n;
		printc(sp, '-');
	}

	for (first = 1; div > 0; div /= base) {
		d = n / div;
		if (d == 0 && first && div != 1 && !fill)
			continue;
		n -= d * div;
		printc(sp, digits[d]);
		first = 0;
	}
}

static long long
getnum(va_list *va, int size)
{
	switch (size) {
	case 0:
		return va_arg(*va, int);
	case 1:
		return va_arg(*va, long);
	case 2:
		return va_arg(*va, long long);
	case 3:
		return (long long) va_arg(*va, void *);
	case 4:
	default:
		return va_arg(*va, size_t);
	}
}

void
doprnt(Stream *sp, const char *fmt, va_list va)
{
	char c;
	va_list va2;
	int fill, base, sign, size;
	char *s;

	va_copy(va2, va);
	while (( c = *fmt++) != '\0') {
		if (c != '%') {
			printc(sp, c);
			continue;
		}

		fill = size = sign = 0;
flags:
		switch (c = *fmt++) {
		case '0':
			fill = 1;
			goto flags;
		case 'l':
			size++;
			goto flags;
		case 'z':
			size = 4;
			goto flags;
		case 'o':
			base = 8;
			goto print_number;
		case 'd':
			sign = 1;
		case 'u':
			base = 10;
			goto print_number;
		case 'p':
			size = 3;
			fill = 1;
		case 'X':
		case 'x':
			base = 16;
		print_number:
			printn(sp, getnum(&va2, size), base, sign, fill);
			break;
		case 'e':
			s = strerror(errno);
			goto printout;
		case 's':
			s = va_arg(va2, char *);
		printout:
			while (*s)
				printc(sp, *s++);
			break;
		case 'c':
			c = va_arg(va2, int);
		case '%':
			printc(sp, c);
			break;
		case '\0':
			goto out_loop;
		}
	}

out_loop:
	va_end(va2);
}
