#include <os9/os9.h>

#include "sysreg.h"
#include "arch.h"

#define NR_LEVELS 7

/* clidr_el1 */
#define LOC_SHIFT       24
#define LOC_MASK       0x7
#define CTYPE_MASK     0x7

/* ccsidr_el1 */
#define NSETS_SHIFT     13
#define NSETS_MASK  0x7fff
#define NWAYS_SHIFT      3
#define NWAYS_MASK   0x3ff
#define LINE_MASK      0x7

/* csselr_el1 */
#define LEVEL_SHIFT      1


static unsigned
log2(unsigned long long v, unsigned width)
{
	unsigned nbits, i = width;

	for (nbits = 0; i-- > 0; nbits++) {
		if ((v & (1ull << i)) != 0)
			break;
	}
	return width - nbits - 1;
}

static void
invdcachelvl(unsigned l)
{
	unsigned long long ccsidr;
	unsigned nsets, nways, size, shift, i, j;

	syswr(CSSELR_EL1, l << LEVEL_SHIFT);
	barrier(CODE);

	ccsidr = sysrd(CCSIDR_EL1);
	nsets = ((ccsidr >> NSETS_SHIFT) & NSETS_MASK) + 1;
	nways = ((ccsidr >> NWAYS_SHIFT) & NWAYS_MASK) + 1;
	size = (ccsidr & LINE_MASK) + 4;
	shift = 32 - log2(nways, 32);

	for (i = 0; i < nways; i++) {
		for (j = 0; j < nsets; j++) {
			uintptr_t addr;

			addr = (uintptr_t)i << shift | j << size | l << 1;
			invdcachesetway((void *)addr);
		}
	}
	barrier(DATA);
}

static void
invdcache(void)
{
	unsigned long long clidr;
	unsigned l;
	enum cachetype {
		NOCACHE,
		ICACHE,
		DCACHE,
		HARVARD,
		UNIFIED
	} type;

	barrier(DATA);

	clidr = sysrd(CLIDR_EL1);
	if (((clidr >> LOC_SHIFT) & LOC_MASK) == 0)
		return;

	for (l = 0; l < NR_LEVELS; l++) {
		type = clidr & CTYPE_MASK;
		if (type == NOCACHE || type == ICACHE)
			continue;

		invdcachelvl(l);
		clidr >>= 3;
	}
}

void
icache(void)
{
	uint64_t  sctlr;

	invicache();
	invdcache();

	sctlr = sysrd(SCTLR_EL1);
	sctlr |= C | I;
	syswr(SCTLR_EL1, sctlr);
}
