#include <libk.h>
#include <os9/os9.h>

#include "arch.h"

static char  regnames[][8] = {
	[X0] = "x0",
	[X1] = "x1",
	[X2] = "x2",
	[X3] = "x3",
	[X4] = "x4",
	[X5] = "x5",
	[X6] = "x6",
	[X7] = "x7",
	[X8] = "x8",
	[X9] = "x9",
	[X10] = "x10",
	[X11] = "x11",
	[X12] = "x12",
	[X13] = "x13",
	[X14] = "x14",
	[X15] = "x15",
	[X16] = "x16",
	[X17] = "x17",
	[X18] = "x18",
	[X19] = "x19",
	[X20] = "x20",
	[X21] = "x21",
	[X22] = "x22",
	[X23] = "x23",
	[X24] = "x24",
	[X25] = "x25",
	[X26] = "x26",
	[X27] = "x27",
	[X28] = "x28",
	[X29] = "x29",
	[ELR] = "elr",
	[X30] = "x30",
	[SPSR] = "spsr",
	[ESR] = "esr",
	[SP] = "sp",
	[FAR] = "far",
	[TTBR0] = "ttbr0",
	[TTBR1] = "ttbr1",
};

void
dumpregs(Context *ctx)
{
	int i;

	for (i = 0; i < NR_REGS; i++) {
		kprint("%s=0x%llx%c", regnames[i], ctx->r[i],
		       (i > 0 && i % 3 == 0) ? '\n' : ' ');
	}
	if (i % 4 != 3)
		kprint("\n");
}

static void
backtrace(Context *ctx)
{
	void **bp;

	kprint("backtrace:\n");
	kprint("%0llx\n", ctx->r[X30]);
	for (bp = (void **) ctx->r[X29]; *bp; bp = (void **) *bp)
		kprint("%p\n", bp[1]);
}

static void
dumpstack(Context *ctx)
{
	long long *sp, i;

	kprint("stack dump:\n");
	sp = (long long *) ctx->r[SP];
	for (i = 1; i <= 16; i++)
		kprint("%0llx%c", *sp++, (i % 4 == 0) ? '\n' : ' ');
}

void
fault(const char *msg, Context *ctx)
{
	static int first = 1;
	Context local;

	if (first) {
		first = 0;
		kprint("panic: %s\n", msg);
		if (!ctx)
			ctx = getctx(&local);
		dumpregs(ctx);
		backtrace(ctx);
		dumpstack(ctx);
	}
	halt();
}

void
panic(const char *msg)
{
	fault(msg, NULL);
}
