#include <os9/os9.h>

#include <errno.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include <deserialize.h>

#include "dev.h"
#include "fip.h"

#define NR_FIPS     2

#define STOC_HEADER_NAME     4               /* strlen(TOC_HEADER_NAME); */
#define STOC_HEADER          16              /* sizeof(fip_toc_header_t); */
#define STOC_ENTRY           40              /* sizeof(fip_toc_entry_t); */

struct fipfile {
	Chan *c;
	long offset[NR_FILES];
	long size[NR_FILES];
};

struct fipuuid {
	char time_low[4];
	char time_mid[2];
	char time_hi_and_version[2];
	char clock_seq_hi_and_reserved;
	char clock_seq_low;
	char node[_UUID_NODE_LEN];
};

struct fipntry {
	struct fipuuid uuid;
	long long      offset_address;
	long long      size;
	long long      flags;
};

static struct fipfile archives[NR_FIPS];
static int nfips;


static int
getntry(Chan *c, struct fipntry *ntry)
{
	int i, n;
	static unsigned char buf[STOC_ENTRY];

	n = devtab[c->type]->read(c, buf, sizeof(buf));
	if (n <= 0)
		return n;

	if (n != sizeof(buf)) {
		errno = EINVAL;
		return -1;
	}

	n = 0;
	for (i = 0; i < 4; i++)
		CHAR(ntry->uuid.time_low[i], buf, n, sizeof(buf));
	for (i = 0; i < 2; i++)
		CHAR(ntry->uuid.time_mid[i], buf, n, sizeof(buf));
	for (i = 0; i < 2; i++)
		CHAR(ntry->uuid.time_hi_and_version[i], buf, n, sizeof(buf));
	CHAR(ntry->uuid.clock_seq_hi_and_reserved, buf, n, sizeof(buf));
	CHAR(ntry->uuid.clock_seq_low, buf, n, sizeof(buf));
	for (i = 0; i < _UUID_NODE_LEN; i++)
		CHAR(ntry->uuid.node[i], buf, n, sizeof(buf));
	LLONG(ntry->offset_address, buf, n, sizeof(buf));
	LLONG(ntry->size, buf, n, sizeof(buf));
	LLONG(ntry->flags, buf, n, sizeof(buf));

	if (ntry->size > LONG_MAX || ntry->offset_address > LONG_MAX) {
		errno = EINVAL;
		return -1;
	}

	if (ntry->size == 0)
		return 0;

	return 1;
}

static int
uuideq(struct fipuuid *u1, struct fipuuid *u2)
{
	if (u1->clock_seq_hi_and_reserved != u2->clock_seq_hi_and_reserved)
		return 0;
	if (u1->clock_seq_low != u2->clock_seq_low)
		return 0;
	if (memcmp(u1->node, u2->node, sizeof(u1->node)))
		return 0;
	if (memcmp(u1->time_hi_and_version, u2->time_hi_and_version, sizeof(u1->time_hi_and_version)))
		return 0;
	if (memcmp(u1->time_low, u2->time_low, sizeof(u1->time_low)))
		return 0;
	if (memcmp(u1->time_mid, u2->time_mid, sizeof(u1->time_mid)))
		return 0;
	return 1;
}

static int
fipgen(Chan *c, Dirtab *tab, int ntab, int n, Dir *dir)
{
	int i, r;
	long off;
	Chan nc;
	struct fipntry ntry;
	struct fipfile *fip;
	static char unk[NAMELEN];

	struct uuidnames {
		char name[NAMELEN];
		struct fipuuid uuid;
	};

	// TODO: generate this array from a text file
	static struct uuidnames uuidnames[] = {
		{"scp-bl2u.bin", UUID_TRUSTED_UPDATE_FIRMWARE_SCP_BL2U},
		{"bl2u.bin", UUID_TRUSTED_UPDATE_FIRMWARE_BL2U},
		{"ns-bl2u.bin", UUID_TRUSTED_UPDATE_FIRMWARE_NS_BL2U},
		{"fwu.crt", UUID_TRUSTED_FWU_CERT},
		{"bl2.bin", UUID_TRUSTED_BOOT_FIRMWARE_BL2},
		{"scp-bl2.bin", UUID_SCP_FIRMWARE_SCP_BL2},
		{"bl31.bin", UUID_EL3_RUNTIME_FIRMWARE_BL31},
		{"bl32.bin", UUID_SECURE_PAYLOAD_BL32},
		{"bl32-xtr1.bin", UUID_SECURE_PAYLOAD_BL32_EXTRA1},
		{"bl32-xtr2.bin", UUID_SECURE_PAYLOAD_BL32_EXTRA2},
		{"bl33.bin", UUID_NON_TRUSTED_FIRMWARE_BL33},
		{"rot-k.crt", UUID_ROT_KEY_CERT},
		{"trstd-k.crt", UUID_TRUSTED_KEY_CERT},
		{"nt-k.crt", UUID_NON_TRUSTED_WORLD_KEY_CERT},
		{"scp-fw-k.crt", UUID_SCP_FW_KEY_CERT},
		{"soc-fw-k.crt", UUID_SOC_FW_KEY_CERT},
		{"tos-fw-k.crt", UUID_TRUSTED_OS_FW_KEY_CERT},
		{"nt-fw-k.crt", UUID_NON_TRUSTED_FW_KEY_CERT},
		{"tb-fw.crt", UUID_TRUSTED_BOOT_FW_CERT},
		{"scp-fw-c.crt", UUID_SCP_FW_CONTENT_CERT},
		{"soc-fw-c.crt", UUID_SOC_FW_CONTENT_CERT},
		{"tos-fw-c.crt", UUID_TRUSTED_OS_FW_CONTENT_CERT},
		{"nt-fw-c.crt", UUID_NON_TRUSTED_FW_CONTENT_CERT},
		{"hw.cfg", UUID_HW_CONFIG},
		{"tb-fw.cfg", UUID_TB_FW_CONFIG},
		{"soc-fw.cfg", UUID_SOC_FW_CONFIG},
		{"tos-fw.cfg", UUID_TOS_FW_CONFIG},
		{"nt-fw.cfg", UUID_NT_FW_CONFIG}
	};

	if (c->dev >= nfips)
		panic("fipgen");
	clone(archives[c->dev].c, &nc);
	fip = &archives[nc.dev];

	off = STOC_HEADER;
	for (i = 0; i <= n; i++) {
		if (fip->offset[i] == -1)
			return 0;
		if (devtab[nc.type]->seek(&nc, off, SEEK_SET) < 0)
			return -1;
		if ((r = getntry(&nc, &ntry)) <= 0)
			return r;
		off += sizeof(ntry);
	}

	for (i = 0; i < NELEM(uuidnames); i++) {
		if (uuideq(&uuidnames[i].uuid, &ntry.uuid))
			break;
	}

	if (i < NELEM(uuidnames)) {
		mkentry(c, dir,
		        uuidnames[i].name, ntry.size,
		        QID(CHFILE, 0, n), O_READ);
	} else {
		// TODO: set name depending on uuid node value
		mkentry(c, dir,
		        unk, ntry.size,
		        QID(CHFILE, 0, n), O_READ);
	}

	return 1;
}

static int
fipwalk(Chan *c, char *name)
{
	return devwalk(c, name, NULL, 0, fipgen);
}

static int
fipstat(Chan *c, char *file, unsigned char *buf, int n)
{
	return devstat(c, file, buf, n, NULL, 0, fipgen);
}

static int
fipread(Chan *c, void *buf, int n)
{
	long off;
	Chan cs;
	struct fipfile *fip;
	long size;
	long long path;

	if (c->qid.type == CHDIR)
		return dirread(c, buf, n, NULL, 0, fipgen);

	if (c->dev >= nfips)
		panic("fipread1");
	fip = &archives[c->dev];

	path = c->qid.path;
	if (path >= NR_FILES || fip->offset[path] < 0)
		panic("fipread2");
	clone(fip->c, &cs);

	size = fip->size[path];
	if (c->offset >= size)
		return 0;

	if (n < 0) {
		errno = EINVAL;
		return -1;
	}
	if (n > size - c->offset)
		n = size - c->offset;

	off = fip->offset[path] + c->offset;
	if (devtab[cs.type]->seek(&cs, off, SEEK_SET) < 0)
		return -1;

	n = devtab[cs.type]->read(&cs, buf, n);
	if (n > 0)
		c->offset += n;

	return n;
}

static Chan *
fipmount(Chan *c, char *spec)
{
	int r, n, t;
	long name;
	Chan *cspec;
	unsigned char buf[STOC_HEADER_NAME];
	struct fipntry ntry;
	struct fipfile *fip;

	if (nfips == NR_FIPS) {
		errno = ENOMEM;
		return NULL;
	}
	fip = &archives[nfips];

	for (n = 0; n < NR_FILES; n++)
		fip->offset[n] = -1;

	if ((cspec = namec(spec, O_READ)) == NULL)
		return NULL;
	fip->c = cspec;

	r = devtab[cspec->type]->read(cspec, buf, STOC_HEADER_NAME);
	if (r < 0)
		goto err;
	n = 0;
	LONG(name, buf, n, sizeof(buf));
	if (r != STOC_HEADER_NAME || name != TOC_HEADER_NAME) {
		errno = EINVAL;
		goto err;
	}

	t = cspec->type;
	if (devtab[t]->seek(cspec, STOC_HEADER, SEEK_SET) < 0)
		goto err;
	for (n = 0; n < NR_FILES; n++) {
		switch (getntry(cspec, &ntry)) {
		case 0:
			return attach('F', nfips++);
		case -1:
			goto err;
		default:
			fip->offset[n] = ntry.offset_address;
			fip->size[n] = ntry.size;
			break;
		}
	}
	errno = ENOMEM;

err:
	delchan(cspec);
	return NULL;
}

Dev fipdevtab = {
	.id = 'F',
	.stat = fipstat,
	.clone = devclone,
	.attach = devattach,
	.walk = fipwalk,
	.read = fipread,
	.write = deverrwrite,
	.mount = fipmount,
	.seek = devseek,
	.sync = devsync,
};
