#include <os9/os9.h>
#include <libk.h>

#include "dev.h"

#define PATH(b, t) \
	((unsigned long long) (b) << 32 | \
	 (unsigned long long) (t))

#define TYPE(q) (((q.path) >> 32) & 0xFF)
#define TASK(q) ((q.path) & 0xFFFFFFFF)

enum procqid {
	Qprocfs,
	Qtaskdir,

	Qtaskattr,
	Qtaskstatus,
};

static Dirtab tasktab[] = {
	{"attr",  QID(CHFILE, 0, Qtaskattr), 0, O_READ},
	{"status", QID(CHFILE, 0, Qtaskstatus), 0, O_READ},
};

static int
procfsgen(Chan *c, Dirtab *tab, int ntab, int n, Dir *dir)
{
	Task *tp;
	char nam[NAMELEN];

	n = getntask(n, &tp);
	if (n <= 0)
		return 0;

	ksnprint(nam, NAMELEN, "%d", tp->tid);
	mkentry(c, dir,
		nam, 0,
		QID(CHDIR, 0, PATH(Qtaskdir, tp->tid)),
	        O_READ);

	return 1;
}

static int
taskgen(Chan *c, Dirtab *tab, int ntab, int n, Dir *dir)
{
	int tid = TASK(c->qid);

	if (!devgen(c, tasktab, NELEM(tasktab), n, dir)) 
		return 0;
	dir->qid.path = PATH(dir->qid.path, tid);

	return 1;
}

static int
procgen(Chan *c, Dirtab *tab, int ntab, int n, Dir *dir)
{
	switch (TYPE(c->qid)) {
	case Qprocfs:
		return procfsgen(c, tab, ntab, n, dir);
	case Qtaskdir:
		return taskgen(c, tab, ntab, n, dir);
	default:
		return 0;
	}
}

static int
procwalk(Chan *c, char *name)
{
	return devwalk(c, name, NULL, 0, procgen);
}

static int
taskattrread(Chan *c, void *buf, int n)
{
	int len;
	Task *tp;
	char tmp[1024];

	if ((tp = gettask(TASK(c->qid))) == NULL)
		panic("");

	len = ksnprint(tmp,
	               sizeof(tmp),
	               "name: %s\n"
	               "tid: %d\n"
	               "entry: %p\n"
	               "stack size: %zd\n"
	               "base prio: %d\n"
	               "period: %ld\n"
	               "capacity: %ld\n"
	               "deadline: %ld\n",
	               tp->name,
	               tp->tid,
	               tp->entry,
	               tp->stack,
	               tp->prio,
	               tp->period,
	               tp->capacity,
	               tp->deadline);

	if (len == sizeof(tmp))
		panic("temporary buffer too short");
	return buf2chan(c, buf, tmp, n, len);
}

static int
taskstatusread(Chan *c, void *buf, int n)
{
	int len;
	Task *tp;
	char tmp[1024];
	static char *states[] = {
		[TDORMANT] = "dormant",
		[TREADY] = "ready",
		[TRUNNING] = "running",
		[TWAITING] = "waiting",
		[FAULTED] = "faulted",
	};

	if ((tp = gettask(TASK(c->qid))) ==  NULL)
		panic("taskstatusread:gettask");

	len = ksnprint(tmp,
	               sizeof(tmp),
	               "prio: %d\n"
	               "flags: %x\n"
	               "deadline time: %ld\n"
	               "state: %s\n"
	               "affinity: %lx\n"
	               "retained prio: %d\n",
	               tp->prio,
	               tp->flags,
	               tp->deadtime,
	               states[tp->state],
	               tp->affinity,
	               tp->retainprio);

	if (len == sizeof(tmp))
		panic("temporary buffer too short");
	return buf2chan(c, buf, tmp, n, len);
}

static int
procread(Chan *c, void *buf, int n)
{
	switch (TYPE(c->qid)) {
	case Qprocfs:
	case Qtaskdir:
                return dirread(c, buf, n, NULL, 0, procgen);
	case Qtaskattr:
		return taskattrread(c, buf, n);
	case Qtaskstatus:
		return taskstatusread(c, buf, n);
	default:
		panic("procread");
	}
}

static int
procstat(Chan *c, char *file, unsigned char *buf, int n)
{
	return devstat(c, file, buf, n, NULL, 0, procgen);
}

Dev procdevtab = {
        .id = 'p',
        .stat = procstat,
        .clone = devclone,
        .attach = devattach,
        .walk = procwalk,
        .read = procread,
        .write = deverrwrite,
        .mount = deverrmount,
        .seek = devseek,
        .sync = devsync,
};
