#include <os9/os9.h>
#include <libk.h>

#include "dev.h"

enum Orootqid {
	Qroot,
	Qdev,
	Qdevuart0,
	Qdevuart1,
	Qdevuart2,
	Qdevuart3,
	Qdevcons,
	Qdevblk,
	Qarfs,
	Qfip,
	Qproc,
	Qblobs,
	Qblobctl,
};

#include "blobfil.h"
#include "blobs.h"

static Dirtab dirtab[] = {
	{"dev",  QID(CHDIR, 0, Qdev), 0, O_READ},
	{"blobs", QID(CHDIR, 0, Qblobs), 0, O_READ},
	{"arfs", QID(CHDIR, 0, Qarfs), 0, O_READ},
	{"fip", QID(CHDIR, 0, Qfip), 0, O_READ},
	{"proc", QID(CHDIR, 0, Qproc), 0, O_READ},
};

static Dirtab devfstab[] = {
	{"uart0", QID(CHDIR, 0, Qdevuart0), 0, O_READ},
	{"uart1", QID(CHDIR, 0, Qdevuart1), 0, O_READ},
	{"uart2", QID(CHDIR, 0, Qdevuart2), 0, O_READ},
	{"uart3", QID(CHDIR, 0, Qdevuart3), 0, O_READ},
	{"cons", QID(CHDIR, 0, Qdevcons), 0, O_READ},
	{"blk", QID(CHDIR, 0, Qdevblk), 0, O_READ},
};

static int
rootgen(Chan *c, Dirtab *tab, int ntab, int n, Dir *dir)
{
	switch (c->qid.path) {
	case Qroot:
		tab = dirtab;
		ntab = NELEM(dirtab);
		break;
	case Qdev:
		tab = devfstab;
		ntab = NELEM(devfstab);
		break;
	case Qblobs:
		tab = blobtab;
		ntab = NELEM(blobtab);
		break;
	default:
		return -1;
	}

	return devgen(c, tab, ntab, n, dir);
}

static int
rootwalk(Chan *c, char *name)
{
	return devwalk(c, name, NULL, 0, rootgen);
}

static int
blobstatus(Chan *c, void *buf, int n)
{
	int i, len;
	Dirtab *dp;
	char info[NELEM(blobtab) * (NAMELEN + 20)];

	len = 0;
	for (i = 1; i < NELEM(blobtab); i++) {
		dp = &blobtab[i];
		len += ksnprint(info+len,
				   sizeof(info) - len,
				   "%s 0x%p %lu\n",
				   dp->name,
				   dp->data,
				   dp->length);
	}
	if (len == sizeof(info))
		panic("blobstatus");

	return buf2chan(c, buf, info, n, len);
}

static int
rootread(Chan *c, void *buf, int n)
{
	Dirtab *dp;

	if (c->qid.type == CHDIR)
		return dirread(c, buf, n, NULL, 0, rootgen);

	if (c->qid.path == Qblobctl)
		return blobstatus(c, buf, n);

	dp = &blobtab[c->qid.path - Qblobctl];
	return buf2chan(c, buf, dp->data, n, dp->length);
}

static int
rootstat(Chan *c, char *file, unsigned char *buf, int n)
{
	return devstat(c, file, buf, n, NULL, 0, rootgen);
}

Dev rootdevtab = {
	.id = '/',
	.stat = rootstat,
	.clone = devclone,
	.attach = devattach,
	.walk = rootwalk,
	.read = rootread,
	.write = deverrwrite,
	.mount = deverrmount,
	.seek = devseek,
	.sync = devsync,
};
